package mapmaking;



import javax.swing.*;

import java.awt.Graphics;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.swing.JButton;
import javax.swing.JLabel;
//import javax.swing.JSeparator;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileFilter;

/**
 *
 * @author Aksel Kvitberg
 */
public class Main {

	BufferedImage[] tree = new BufferedImage[11];
	BufferedImage[] alpha = new BufferedImage[11];
	BufferedImage mask;
	BufferedImage workingMask;
	BufferedImage workImg;
	final int staticTreeSize = 100; 
//	final int staticGridSize = 60;
        int maxTreeWidth = 0;
        int maxTreeHeight = 0;
        int treeGapX;
        int treeGapY;

	int treeSize = 19;
        int treeHeight;
	int gridsPerTree = 4;
	int gridSize = 10; //Increasing these makes tree futher apart.
        int ygridSize = 10; //
	int width;
	int height;

        int minJitterX = 2;
        int minJitterY = 2;

        int jitterX = 2;  // Variables to hold slight jittering of tree placements.
        int jitterY = 6;

        int treeSpacing = 0; // Variable for the random chance to move trees along on X axis.

	JFrame frame;
	ImageViewer viewer;
	int id = 1;
	private int gridX;
	private int gridY;
	int imageMatrix[][];
	int[][] treeMatrix;
	private boolean maskLoaded = false;
	JMenuBar jmbMenu = new JMenuBar();
	JMenuItem jmSave = new JMenuItem("Save");
	JMenuItem jmLoad = new JMenuItem("Load");
	JFileChooser jfc = new JFileChooser();
	JCheckBox jcbFill = new JCheckBox("Masked");
	JSlider jsScale = new JSlider(10, 200);


        SpinnerModel model1 = new SpinnerNumberModel(10, //initial value
                                           1, //min
                                           20, //max
                                           1);
        JSpinner DensitySpinX = new JSpinner(model1);
        SpinnerModel model2 = new SpinnerNumberModel(10, //initial value
                                           1, //min
                                           20, //max
                                           1);
        JSpinner DensitySpinY = new JSpinner(model2);

        JLabel scaleLabel = new JLabel("Scale", JLabel.CENTER);
        JLabel XdensityLabel = new JLabel("   X-Density: ", JLabel.CENTER);
        JLabel YdensityLabel = new JLabel("   Y-Density: ", JLabel.CENTER);
	JMenuItem jmGenerate = new JMenuItem("Generate");



	public Main() {

		frame = new JFrame("TreeThing");
		viewer = new ImageViewer(this);

		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setContentPane(viewer);
		frame.setBounds(100, 10, 800, 800);
		frame.setVisible(true);

		jfc.setFileFilter(new FileFilter() {

			@Override
			public boolean accept(File f) {
				if (f.isDirectory()) {
					return true;
				}
				String ext = null;
				String s = f.getName();
				int i = s.lastIndexOf('.');

				if (i > 0 && i < s.length() - 1) {
					ext = s.substring(i + 1).toLowerCase();
				}

				String extension = ext;
				if (extension != null) {
					if (extension.equals("png")) {
						return true;
					}
				}
				return false;
			}

			@Override
			public String getDescription() {
				return ".png";
			}
		});
		jfc.setAcceptAllFileFilterUsed(false);

		frame.setJMenuBar(jmbMenu);
		jmbMenu.add(jmLoad);
		jmbMenu.add(jmSave);
		jmbMenu.add(jmGenerate);


//                jmbMenu.add(new JSeparator());

		jmbMenu.add(scaleLabel);
		jmbMenu.add(jsScale);
		jmbMenu.add(XdensityLabel);

                jmbMenu.add(DensitySpinX);
		jmbMenu.add(YdensityLabel);
                jmbMenu.add(DensitySpinY);

		jmbMenu.add(jcbFill);
		jsScale.setValue(20);
                
                changeTreeSize(); //Added this here so that it initialise trees and doesn't error when an element is used.


		jmLoad.addActionListener(
				new ActionListener() {

					public void actionPerformed(ActionEvent e) {
						loadImage();
					}
				});

		jmSave.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				saveImage();
			}
		});
		jmGenerate.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {

				drawTrees();
			}
		});
		jcbFill.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {

				drawTrees();
			}
		});
		jsScale.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				changeTreeSize();
			}
		});


		DensitySpinX.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				changeTreeDensity();
				drawTrees();
			}
		});
		DensitySpinY.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				changeTreeDensity();
				drawTrees();
			}
		});
	}

	public void changeTreeSize() {

		double percent = jsScale.getValue() / 100.0;
		loadTreeImages(percent);
		treeSize = (int) (maxTreeWidth * percent);
                treeHeight = (int) (maxTreeHeight * percent);
	}


	public void changeTreeDensity() {

		gridSize = ((Integer)DensitySpinX.getValue()).intValue();
 		ygridSize = ((Integer)DensitySpinY.getValue()).intValue();

                if (maskLoaded) createImage(); // Recall this function to repopulate the imageMatrix array upon density change.
	}


        public void loadTreeImages(double size) {

		for (int i = 1; i <
				tree.length; i++) {
			tree[i] = readImage("Images/Tree" + i + ".png");
			alpha[i] = readImage("Images/Alpha" + i + ".png");



                        if (tree[i].getWidth() > maxTreeWidth)
                                maxTreeWidth = tree[i].getWidth();
                        if (tree[i].getHeight() > maxTreeHeight)
                                maxTreeHeight = tree[i].getHeight();



			BufferedImageOp op = new AffineTransformOp(
					AffineTransform.getScaleInstance(size, size),
					new RenderingHints(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC));
			tree[i] = op.filter(tree[i], null);
			alpha[i] = op.filter(alpha[i], null);
		}

	}

	public void createImage() {
		width = mask.getWidth();
		height = mask.getHeight();

		BufferedImageOp op = new AffineTransformOp(
				AffineTransform.getScaleInstance(1.0 / (gridSize), 1.0 / (ygridSize)), //Why 0.2? Can't remember. Probably not right.
				new RenderingHints(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC));
		workingMask =
				op.filter(mask, null);
		System.out.println(workingMask.getWidth());
		Raster raster = workingMask.getAlphaRaster();

		imageMatrix =
				getCleanMatrix();
//		System.out.println("rasterWidth: " + raster.getWidth());
		for (int x = 0; x <
				gridX; x++) {
			for (int y = 0; y <
					gridY; y++) {
//				System.out.print(raster.getSample(x, y, 0) + " ");
				if (raster.getSample(x, y, 0) == 0) {
					imageMatrix[x][y] = 255;
				} else {
					imageMatrix[x][y] = 0;
				}

			}
			System.out.println();
		}

		maskLoaded = true;
		drawTrees();

	}

	public int[][] getCleanMatrix() {
		gridX = (width / gridSize);
		gridY =	(height / ygridSize);

		return new int[gridX][gridY];
	}

	public void drawTrees() {
		int[][] matrix = getCleanMatrix();
		if (maskLoaded) {
			for (int x = 0; x <
					gridX; x++) {
				for (int y = 0; y <
						gridY; y++) {
					matrix[x][y] = imageMatrix[x][y];
				}

			}
		} else {
			width = 1024;
			height = 1024;
			matrix = getCleanMatrix();
		}

		workImg = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		viewer.setImage(workImg);
		drawTrees(matrix);
	}

	public void drawTrees(int[][] matrix) {
		treeMatrix = getCleanMatrix();
		Graphics g = workImg.getGraphics();
//		int jumpStep = 3;
//		for (int i = jumpStep; i >
//				0; i--) {
//			generateTrees(treeMatrix, matrix, jumpStep * gridsPerTree);
//		}

		generateTrees(treeMatrix, matrix, gridsPerTree);
		paintTrees();

//		System.out.println("Stopped");
	}


	public void generateTrees(int[][] trees, int[][] matrix, int jump) {
		Random r = new Random();
		for (int y = 0; y < gridY - gridsPerTree; y++) {
                    for (int x = 0; x < gridX - gridsPerTree; x++) {

                        int randomHold = r.nextInt(tree.length);
                        if (randomHold == 0) randomHold = 1;

                        if (checkMatrix(matrix, x, y)) {
                                trees[x][y] = randomHold;
                                fillMatrix(matrix, x, y);
                        }
                    }
		}
	}



	public void paintTrees() {
		Graphics g = workImg.getGraphics();
		Random r = new Random();


                int randX = 0;
                int randY = 0;
                int rowCount = 0;
                int lastXJitter = 0;

		for (int y = 0; y < gridY; y = y + 1) {

                       int gridAdd = 0;
                       int totalXJitter = 0;

                        int rowIndent = r.nextInt(2);
                        if (rowIndent == 1) {
                            rowCount++;
                            if (rowCount > 4) rowCount = 0;
                        }
                        else {
                            rowCount--;
                            if (rowCount < 0) rowCount = 3;
                        }


			for (int x = rowCount; x < gridX; x = x + 4) { //Four grids per X movement now to allow more staggering of the following rows.


                                if (treeSpacing > 0) {
                                    int tempPercent = r.nextInt(treeSpacing);
                                    if (tempPercent < 2) { gridAdd = gridAdd + 1;
                                            x = x + gridAdd;
                                                if (x >= gridX) x = x - gridAdd;
                                    }
                                }


				if (treeMatrix[x][y] != 0) {
					int nextTree = treeMatrix[x][y];
					if (!jcbFill.isSelected()) {



                                           if (ygridSize > (int)(treeHeight / 4))
                                               jitterY = (ygridSize  - (int)(treeHeight / 4)); // This leaves a quarter of the tree still uncovered at the top. Hopefully.

                                           if (treeSize - 2 < (gridSize * 4))
                                                treeGapX = ((gridSize * 4) - treeSize) + (int)(treeSize / 4); //Allow a little overlap to give more jitter room.
                                            else {
                                                    treeGapX = 0;
                                           //         jitterY = 0; //Y shouldn't jitter if X is overlapped as it puts bad overlaps down.
                                            }
                                            if (treeGapX > lastXJitter)
                                                 jitterX = (treeGapX - lastXJitter);
                                            else jitterX = minJitterX;

//System.out.println("treeGapX: " + treeGapX + "        JitterY: " + jitterY);


                                            randX = r.nextInt(jitterX);
                                            lastXJitter = randX;

                                            if (treeGapX == 0) {
                                                        randX = randX + totalXJitter;
                                                        totalXJitter = randX;
                                                        }
                                            if (jitterY > 0) randY = r.nextInt(jitterY);


						Raster alphaRaster = alpha[nextTree].getAlphaRaster();
						WritableRaster raster = workImg.getAlphaRaster();

						int aWidth = alphaRaster.getWidth();
						int aHeight = alphaRaster.getHeight();
						double[] doub = new double[4];
						raster.getPixel(x, y, doub);
						doub[0] = 0;

//						System.out.println("raster: " + alphaRaster.getWidth() + ", " + alphaRaster.getHeight());
						for (int ax = 0; ax <
								aWidth; ax++) {
							for (int ay = 0; ay <
									aHeight; ay++) {
								if (alphaRaster.getSample(ax, ay, 0) > 0) {
									try {
                 						//alphaRaster.getPixel(ax, ay, doub);
                                                                //doub[0] = 0;
										raster.setPixel(x * gridSize + ax + randX, y * ygridSize + ay + randY, doub);
//										System.out.println("Tried to draw " + x * gridSize + ax + ", " + y * gridSize + ay);
									} catch (Exception e) {
//										System.out.println("Error: Tried to draw " + x * gridSize + ax + ", " + y * gridSize + ay);
									}
								}
							}
						}
						g.drawImage(tree[nextTree], x * gridSize + randX, y * ygridSize - randY, null);
					} else {
                                                if (jitterX > 0) randX = r.nextInt(jitterX);
                                                if (jitterY > 0) randY = r.nextInt(jitterY);
						g.drawImage(alpha[nextTree], x * gridSize + randX, y * ygridSize + randY, null);
						g.drawImage(tree[nextTree], x * gridSize + randX, y * ygridSize + randY, null);
					}

				}
			}
		}
		viewer.setImage(workImg);
	}


        public void yJitterCheck() {

//int treeGap;
//int lastYJitter = 0;
//int randY;

        }

	public boolean checkMatrix(int m[][], int x, int y) {
		for (int i = x + 1; i < x + gridsPerTree - 1; i++) {
			for (int j = y; j < y + gridsPerTree - 3; j++) {
				if (m[i][j] != 0) {
					return false;
				}

			}
		}

		return true;
	}

	public void fillMatrix(int m[][], int x, int y) {

		for (int i = x; i <
				gridsPerTree; i++) { // -1 allows no overlap on X.
			for (int j = y; j <
					gridsPerTree; j++) { //-3 allows 50% overlap on Y.

				m[i][j] = 1;
			}

		}
	}

	public void printMatrix(int m[][]) {
		for (int m1[] : m) {
			for (int m2 : m1) {
				System.out.print(m2 + " ");
			}

			System.out.println("");
		}

	}

	void updateImage(int sm, double yS, double xS, double yG, double xG, double r, boolean tr) {
		drawTrees();
	}

	public BufferedImage readImage(
			String path) {
		try {
			BufferedImage img = ImageIO.read(new File(path));
			return img;
		} catch (IOException ex) {
			System.err.println("Error 1. Cannot load image");
			Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
			System.exit(
					0);









			return null;
		}
	}

	public void loadImage() {
		int returnVal = jfc.showOpenDialog(frame);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			try {
				File f = jfc.getSelectedFile();
				mask =
						ImageIO.read(f);
				createImage();

			} catch (IOException ex) {
				System.err.println("Error 1. Cannot load image");
				Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
			}
		}

	}

	public void saveImage() {
		int returnVal = jfc.showSaveDialog(frame);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File f = jfc.getSelectedFile();
			if (f.exists()) {
				int ans = JOptionPane.showConfirmDialog(frame, "File " + f.getName() + " exists already. Overwrite?", "Save", JOptionPane.YES_NO_OPTION);
				if (ans == JOptionPane.OK_OPTION) {
					saveImage(f);
				}

			}
			saveImage(f);
		}

	}

	public void saveImage(File f) {
		try {
			String ext = null;
			String s = f.getName();
			int i = s.lastIndexOf('.');

			if (i > 0 && i < s.length() - 1) {
				ext = s.substring(i + 1).toLowerCase();
			}

			if (ext != null) {
				ImageIO.write(workImg, "PNG", f);
			} else{
				File fil = new File(f.getAbsolutePath() + ".png");
				ImageIO.write(workImg, "PNG", fil);
			}
		} catch (IOException ex) {
			Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	/**
	 * @param args the command line arguments
	 */
	public static void main(String[] args) {
		new Main();
	}
}
